import socket

from .stream_receiver import StreamReceiver
from gbvision.exceptions.tcp_stream_closed import TCPStreamClosed


class TCPStreamReceiver(StreamReceiver):
    """
    this class uses TCP to receive a stream over the network, the stream is by default set to be MJPEG
    the broadcaster is the server and the receiver is the client

    :param ip: the IPv4 address of the stream broadcaster, for example '10.45.90.8'
    :param port: the port which TCP should use
    """

    def __init__(self, ip: str, port: int, *args, **kwargs):
        StreamReceiver.__init__(self, *args, **kwargs)
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.server_addr = (ip, port)
        self.socket.connect(self.server_addr)

    def _receive(self):
        try:
            return self.socket.recv(4096)
        except OSError as e:
            raise TCPStreamClosed() from e




import numpy as np
from copy import deepcopy

from gbvision.constants.types import Number


class CameraData:
    """
    describes constant about a camera in it's default state used to approximate distance
    between the camera and an object seen in a frame

    :param focal_length: the focal length of the camera at it's default state, in units of pixels
        can be described as the square root of the amount of pixels an object takes on a frame, multiplied by it's
        distance from the camera and divided by the square root of it's surface

        FOCAL_LENGTH = :math:' sqrt(P) * D / sqrt(S)'

        where P is the amount of pixels in the frame representing the object,
        D is the real life distance between the object and the camera
        S is the real life surface area (in 2d projection) of the object
        note that this is a constant, whatever object you choose to use, this formula will yield the same result
    :param fov:
        half the viewing angle of the camera (field of view) in radians, can be calculated by placing an object in front
        of the camera, so that the entire object is captured and it's center is at the frame's center.
        the tangent of the angle can be described as the width of the object in real life, divided by the
        product of the object's distance from the camera in real life and the ratio between the width of the frame
        in pixels and the width of the object in the frame, also in pixels

        :math:'tan(FOV) = (Wm) / (D * (Wp/Wf))'

        where Wm is the real life width of the object
        D is the real life distance between the object and the camera
        Wp is the width of the object in the frame (pixels unit)
        Wf is the width of the frame (pixels unit)
        to calculate the FOV just apply the inverse tangent

        FOV = :math:'arctan(tan(FOV))'

    :param yaw_angle:
        the clockwise yaw angle (in radians) in which the camera is rotated, the yaw angle is the angle around the y axis,
        it's output only affects the x and z axises.
        set this variable when the camera is rotated around the y axis and you want the output of finder functions
        to represent the original space, rather then the rotated one.
    :param pitch_angle:
        the clockwise pitch angle (in radians) in which the camera is rotated, the pitch angle is the angle around the x axis,
        it's output only affects the y and z axises.
        set this variable when the camera is rotated around the x axis and you want the output of finder functions
        to represent the original space, rather then the rotated one.
    :param roll_angle:
        the clockwise roll angle (in radians) in which the camera is rotated, the roll angle is the angle around the z axis,
        it's output only affects the x and y axises.
        set this variable when the camera is rotated around the z axis and you want the output of finder functions
        to represent the original space, rather then the rotated one.
    :param x_offset:
        the x offset in which the camera is placed
        the distance from the measuring point (usually the center of the robot) to the camera on the x axis (left/right),
        if the camera is to the right this should be positive and if it is left this should be negative
    :param y_offset:
        the y offset in which the camera is placed
        the distance from the measuring point to the camera on the y axis (up/down), if the camera is above the measuring point
        this variable should be positive and if it is below this should be negative
    :param z_offset:
        the z offset in which the camera is placed
        the distance from the measuring point to the camera on the z axis (depth), if the camera is placed outer then the measuring point
        this variable should be positive and if it is inner this should be negative
    :param is_immutable: determines whether the camera data object's values are immutable (True) or mutable (False)
    """

    def __init__(self, focal_length, fov, pitch_angle=0, yaw_angle=0, roll_angle=0, x_offset=0, y_offset=0, z_offset=0,
                 is_immutable=False, name=None):
        self.focal_length = focal_length
        self.fov = fov
        self.rotation_angles = np.array([pitch_angle, yaw_angle, roll_angle])
        self.rotation_matrix = self.__calculate_rotation_matrix()
        self.offset = np.array([x_offset, y_offset, z_offset])
        self.name = name
        self.__is_immutable = is_immutable

    def __calculate_rotation_matrix(self):

        pitch_angle, yaw_angle, roll_angle = self.rotation_angles
        sin, cos = np.sin(yaw_angle), np.cos(yaw_angle)
        rotation_matrix = np.array([[cos, 0, sin],
                                    [0, 1, 0],
                                    [-sin, 0, cos]])
        sin, cos = np.sin(pitch_angle), np.cos(pitch_angle)
        rotation_matrix = rotation_matrix.dot(np.array([[1, 0, 0],
                                                        [0, cos, -sin],
                                                        [0, sin, cos]]))
        sin, cos = np.sin(roll_angle), np.cos(roll_angle)
        rotation_matrix = rotation_matrix.dot(np.array([[cos, -sin, 0],
                                                        [sin, cos, 0],
                                                        [0, 0, 1]]))
        return rotation_matrix

    def __get_data(self) -> 'CameraData':
        return self.copy() if self.__is_immutable else self

    def rotate_pitch(self, angle: float) -> 'CameraData':
        """
        rotates the camera's angle around the pitch axis (the x axis)

        :param angle: the rotation angle
        :return: a camera data instance with the same params as this but with the pitch angle rotated \
            if this is immutable it will return a copy of this, otherwise it will modify this instance and return it
        """
        data = self.__get_data()
        data.rotation_angles[0] += angle
        sin, cos = np.sin(angle), np.cos(angle)
        data.rotation_matrix = data.rotation_matrix.dot(np.array([[1, 0, 0],
                                                                  [0, cos, -sin],
                                                                  [0, sin, cos]]))
        return data

    def rotate_yaw(self, angle: float) -> 'CameraData':
        """
        rotates the camera's angle around the yaw axis (the y axis)

        :param angle: the rotation angle
        :return: a camera data instance with the same params as this but with the yaw angle rotated \
            if this is immutable it will return a copy of this, otherwise it will modify this instance and return it
        """
        data = self.__get_data()
        data.rotation_angles[1] += angle
        sin, cos = np.sin(angle), np.cos(angle)
        data.rotation_matrix = data.rotation_matrix.dot(np.array([[cos, 0, sin],
                                                                  [0, 1, 0],
                                                                  [-sin, 0, cos]]))
        return data

    def rotate_roll(self, angle: float) -> 'CameraData':
        """
        rotates the camera's angle around the roll axis (the z axis)

        :param angle: the rotation angle
        :return: a camera data instance with the same params as this but with the roll angle rotated \
            if this is immutable it will return a copy of this, otherwise it will modify this instance and return it
        """
        data = self.__get_data()
        data.rotation_angles[2] += angle
        sin, cos = np.sin(angle), np.cos(angle)
        data.rotation_matrix = data.rotation_matrix.dot(np.array([[cos, -sin, 0],
                                                                  [sin, cos, 0],
                                                                  [0, 0, 1]]))
        return data

    def set_pitch_angle(self, angle: float) -> 'CameraData':
        """
        sets the camera's angle around the pitch axis (the x axis)

        :param angle: the rotation angle
        :return: a camera data instance with the same params as this but with the pitch angle changed \
            if this is immutable it will return a copy of this, otherwise it will modify this instance and return it
        """
        data = self.__get_data()
        data.rotation_angles[0] = angle
        data.rotation_matrix = data.__calculate_rotation_matrix()
        return data

    def set_yaw_angle(self, angle: float) -> 'CameraData':
        """
        sets the camera's angle around the yaw axis (the y axis)

        :param angle: the rotation angle
        :return: a camera data instance with the same params as this but with the yaw angle changed \
            if this is immutable it will return a copy of this, otherwise it will modify this instance and return it
        """
        data = self.__get_data()
        data.rotation_angles[1] = angle
        data.rotation_matrix = data.__calculate_rotation_matrix()
        return data

    def set_roll_angle(self, angle: float) -> 'CameraData':
        """
        sets the camera's angle around the roll axis (the z axis)

        :param angle: the rotation angle
        :return: a camera data instance with the same params as this but with the roll angle changed \
            if this is immutable it will return a copy of this, otherwise it will modify this instance and return it
        """
        data = self.__get_data()
        data.rotation_angles[2] = angle
        data.rotation_matrix = data.__calculate_rotation_matrix()
        return data

    def move_x(self, x: Number) -> 'CameraData':
        """
        moves this camera data's x axis offset

        :param x: the x offset to move by
        :return: a camera data instance with the same params as this but with the x axis moved \
            if this is immutable it will return a copy of this, otherwise it will modify this instance and return it
        """
        data = self.__get_data()
        data.offset[0] += x
        return data

    def move_y(self, y: Number) -> 'CameraData':
        """
        moves this camera data's y axis offset

        :param y: the y offset to move by
        :return: a camera data instance with the same params as this but with the y axis moved \
            if this is immutable it will return a copy of this, otherwise it will modify this instance and return it
        """
        data = self.__get_data()
        data.offset[1] += y
        return data

    def move_z(self, z: Number) -> 'CameraData':
        """
        moves this camera data's z axis offset

        :param z: the z offset to move by
        :return: a camera data instance with the same params as this but with the z axis moved \
            if this is immutable it will return a copy of this, otherwise it will modify this instance and return it
        """
        data = self.__get_data()
        data.offset[2] += z
        return data

    def set_x_offset(self, x: Number) -> 'CameraData':
        """
        sets this camera data's x axis offset

        :param x: the new x offset
        :return: a camera data instance with the same params as this but with the x axis changed \
            if this is immutable it will return a copy of this, otherwise it will modify this instance and return it
        """
        data = self.__get_data()
        data.offset[0] = x
        return data

    def set_y_offset(self, y: Number) -> 'CameraData':
        """
        sets this camera data's y axis offset

        :param y: the new y offset
        :return: a camera data instance with the same params as this but with the y axis changed \
            if this is immutable it will return a copy of this, otherwise it will modify this instance and return it
        """
        data = self.__get_data()
        data.offset[1] = y
        return data

    def set_z_offset(self, z: Number) -> 'CameraData':
        """
        sets this camera data's z axis offset

        :param z: the new z offset
        :return: a camera data instance with the same params as this but with the z axis changed \
            if this is immutable it will return a copy of this, otherwise it will modify this instance and return it
        """
        data = self.__get_data()
        data.offset[2] = z
        return data

    def copy(self) -> 'CameraData':
        """
        creates a mutable copy of this and returns it
        :return:
        """
        copy = deepcopy(self)
        copy.__is_immutable = False
        return copy

    def is_immutable(self) -> bool:
        """
        checks if this camera data instance is immutable

        :return: True if this is immutable, False otherwise
        """
        return self.__is_immutable

    def as_immutable(self) -> 'CameraData':
        """
        creates and returns an immutable copy of this camera data
        if this instance is already immutable it will return this instance

        :return: an instance of CameraData, with the same values as this instance but immutable
        """
        if self.__is_immutable:
            return self
        copy = self.copy()
        copy.__is_immutable = True
        return copy

    def __str__(self):
        if self.name is not None:
            return f'{self.name}'
        return object.__str__(self)

    def __repr__(self):
        return str(self)



import cv2
import numpy as np

from gbvision.utils.pipeline import PipeLine


@PipeLine
def corners(frame):
    return cv2.filter2D(frame, -1, np.array([[-1, 1], [1, -1]]))


@PipeLine
def edges(frame):
    return cv2.filter2D(frame, -1, np.array([[-1, -1, -1], [-1, 8, -1], [-1, -1, -1]]))


@PipeLine
def sharpen(frame):
    return cv2.filter2D(frame, -1, np.array([[-1, -1, -1], [-1, 9, -1], [-1, -1, -1]]))


@PipeLine
def blur(frame):
    return cv2.filter2D(frame, -1, np.array([[1, 1, 1], [1, 1, 1], [1, 1, 1]]) / 9)


@PipeLine
def blue(frame):
    return frame[:, :, 0]


@PipeLine
def green(frame):
    return frame[:, :, 1]


@PipeLine
def red(frame):
    return frame[:, :, 2]


@PipeLine
def gray(frame):
    return cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)



from typing import Union, List, Generator, Any, Tuple

from .camera_data import CameraData
from .stream_camera import Camera, StreamCamera
from gbvision.constants.types import Frame


class CameraList(Camera):
    """
    behaves as both a camera and a list of cameras
    camera list holds in it a list of cameras referenced as the field cameras
    and also a single camera to be the current camera used for every operation on the camera list
    as a single camera

    :param cameras: list of the cameras which will be part of the camera list
        you can also add and remove cameras later using the
    :param select_cam: optional, an initial camera to be selected, if not set default camera is the first
        one in the list
    """

    def __init__(self, cameras: List[Camera], select_cam: int = None):
        self.cameras: List[Camera] = cameras.copy()
        if select_cam is None and len(cameras) > 0:
            select_cam = 0
        self.selected_camera: Union[Camera, StreamCamera] = self.cameras[select_cam] if select_cam is not None else None

    def __getitem__(self, item: int) -> Camera:
        """
        returns the camera at the index
        :param item: the index
        :return: the camera
        """
        return self.cameras[item]

    def __setitem__(self, item: int, value: Camera):
        """
        sets the camera at the index to the new camera
        :param item: the index
        :param value: the new camera
        """
        self.cameras[item] = value

    def select_camera(self, index: int):
        """
        sets the selected camera to be the camera at the index
        :param index: the new selected camera's index
        """
        self.selected_camera = self.cameras[index]

    def __delitem__(self, item: int):
        """
        deletes the camera at the index
        :param item:
        """
        if self.selected_camera is self.cameras[item]:
            self.selected_camera = None
        del self.cameras[item]

    def __iter__(self):
        """
        :return: an iterator that iterates through all the cameras
        """
        return iter(self.cameras)

    def __len__(self):
        return len(self.cameras)

    def read(self, foreach=False) -> Union[Tuple[bool, Frame], Generator[Tuple[bool, Frame], Any, None]]:
        if foreach:
            return (cam.read() for cam in self.cameras)
        return self.selected_camera.read()

    def is_opened(self, foreach=False) -> Union[bool, Generator[bool, Any, None]]:
        if foreach:
            return (cam.is_opened() for cam in self.cameras)
        return self.selected_camera.is_opened()

    def add_camera(self, cam: Camera):
        """
        adds a new camera to the end of the list
        :param cam: the new camera
        """
        self.cameras.append(cam)

    def release(self, foreach=False):
        if foreach:
            for cam in self.cameras:
                cam.release()
        else:
            self.selected_camera.release()
            self.selected_camera = None

    def default(self):
        """
        sets the selected camera to the default camera (first camera in the list)
        """
        self.selected_camera = self.cameras[0] if len(self.cameras) > 0 else None

    def set_exposure(self, exposure, foreach=False) -> Union[bool, Generator[bool, Any, None]]:
        if foreach:
            return (cam.set_exposure(exposure) for cam in self.cameras)
        else:
            return self.selected_camera.set_exposure(exposure)

    def set_auto_exposure(self, auto, foreach=False) -> Union[bool, Generator[bool, Any, None]]:
        if foreach:
            return (cam.set_auto_exposure(auto) for cam in self.cameras)
        else:
            return self.selected_camera.set_auto_exposure(auto)

    def get_data(self, foreach=False) -> Union[CameraData, Generator[CameraData, Any, None]]:
        if foreach:
            return (cam.get_data() for cam in self.cameras)
        return self.selected_camera.get_data()

    def resize(self, x_factor, y_factor, foreach=False):
        if foreach:
            for cam in self.cameras:
                cam.resize(x_factor, y_factor)
        else:
            self.selected_camera.resize(x_factor, y_factor)

    def rescale(self, factor, foreach=False):
        if foreach:
            for cam in self.cameras:
                cam.rescale(factor)
        else:
            self.selected_camera.rescale(factor)

    def set_frame_size(self, width, height, foreach=False):
        if foreach:
            for cam in self.cameras:
                cam.set_frame_size(width, height)
        else:
            self.selected_camera.set_frame_size(width, height)

    def toggle_stream(self, should_stream, foreach=False):
        if foreach:
            for cam in self.cameras:
                if isinstance(cam, StreamCamera):
                    cam.toggle_stream(should_stream)
        else:
            self.selected_camera.toggle_stream(should_stream)

    def is_streaming(self, foreach=False) -> Union[bool, Generator[bool, Any, None]]:
        if foreach:
            return (cam.is_streaming() if isinstance(cam, StreamCamera) else False for cam in self.cameras)
        return self.selected_camera.is_streaming()

    def get_width(self, foreach=False) -> Union[int, Generator[int, Any, None]]:
        if foreach:
            return (cam.get_width() for cam in self.cameras)
        return self.selected_camera.get_width()

    def get_height(self, foreach=False) -> Union[int, Generator[int, Any, None]]:
        if foreach:
            return (cam.get_height() for cam in self.cameras)
        return self.selected_camera.get_height()

    def _set_width(self, width: int, foreach=False):
        if foreach:
            for cam in self.cameras:
                cam._set_width(width)
        else:
            self.selected_camera._set_width(width)

    def _set_height(self, height: int, foreach=False):
        if foreach:
            for cam in self.cameras:
                cam._set_height(height)
        else:
            self.selected_camera._set_height(height)



from typing import Tuple

from gbvision.constants.types import Frame
from .stream_camera import SimpleStreamCamera
from gbvision.models.cameras import UNKNOWN_CAMERA
from .usb_camera import USBCamera


class USBStreamCamera(SimpleStreamCamera, USBCamera):
    """
    a simple USB stream camera
    """

    def _read(self) -> Tuple[bool, Frame]:
        return USBCamera.read(self)

    def __init__(self, broadcaster, port, should_stream=False, data=UNKNOWN_CAMERA):
        SimpleStreamCamera.__init__(self, broadcaster, should_stream=should_stream)
        USBCamera.__init__(self, port, data)



from typing import List, Dict, Callable, Union

from gbvision.constants.types import Frame, Shape
from gbvision.utils.continuity import ContinuesCircle
from gbvision.utils.continuity.continues_rect import ContinuesRect
from gbvision.utils.continuity import ContinuesRotatedRect
from gbvision.utils.continuity import ContinuesShape
from gbvision.utils.tracker import Tracker

_CONTINUES_SHAPE_TYPES = {
    'CIRCLE': ContinuesCircle,
    'RECT': ContinuesRect,
    'ROTATED_RECT': ContinuesRotatedRect

}


class ContinuesShapeWrapper:
    """
    an object that tracks several shapes in a frame using continuity

    :param shapes: a list of shapes to track using continuity (must be of the same shape)
    :param frame: the frame from which the shapes were found
    :param finding_pipeline: a function that finds the shapes in a given frame and returns a list of them (order irrelevant)
    :param shape_type: the type of the shape, can be either 'CIRCLE', 'RECT', or 'ROTATED_RECT', default is 'RECT', can also be a class that inherits from ContinuesShape
    :param tracker_type: the type of the trackers to use, default is 'EMPTY'
    :param shape_lifespan: the maximum amount of frames for a shape to not be found until it is considered lost
    :param track_new: indicates whether to track new shapes that were un-tracked so far or ignore them, default is False (ignore)
    :param args: additional arguments for continues shape constructor 
    :param kwargs: additional keyword arguments for continues shape constructor
    """

    SHAPE_TYPE_CIRCLE = 'CIRCLE'
    SHAPE_TYPE_RECT = 'RECT'
    SHAPE_TYPE_ROTATED_RECT = 'ROTATED_RECT'

    def __init__(self, shapes: List[Shape], frame: Frame, finding_pipeline: Callable[[Frame], List[Shape]],
                 shape_type: Union[str, type] = 'RECT', tracker_type=Tracker.TRACKER_TYPE_EMPTY,
                 shape_lifespan: int = None, track_new=False, *args, **kwargs):

        if shape_type in _CONTINUES_SHAPE_TYPES:
            shape_type = shape_type.upper()
            self.shape_type = _CONTINUES_SHAPE_TYPES[shape_type]
        else:
            self.shape_type = shape_type
        self.tracker_type = tracker_type
        self.shape_lifespan = shape_lifespan
        self.finding_pipeline = finding_pipeline
        self.shapes: Dict[int, ContinuesShape] = {}
        self.track_new = track_new
        self.__args = args
        self.__kwargs = kwargs
        for i, shape in enumerate(shapes):
            self.shapes[i] = self.__create_continues_shape(shape, frame)
        self.__idx = len(shapes)

    def __create_continues_shape(self, shape, frame) -> ContinuesShape:
        return self.shape_type(shape, frame, Tracker(self.tracker_type), *self.__args, **self.__kwargs)

    def find_shapes(self, frame: Frame) -> Dict[int, Shape]:
        """
        finds all shapes in the frame, them performs a continues shape operations on them and return the result as a dict where the keys are unique ids and the values are the shapes
        if a shape was lost it removes it from the tracked shapes list
        if a new shape was found and the track_new field is set to True it adds it to the tracked shapes list

        :param frame: the frame to search in
        :return: a dict mapping from unique ids to shapes, based on continuity
        """
        shapes = self.finding_pipeline(frame)
        result = {}
        to_delete = []
        for i in self.shapes:
            cont_shape = self.shapes[i]
            if cont_shape.is_lost(self.shape_lifespan):
                to_delete.append(i)
                continue
            found = False
            for j, shape in enumerate(shapes):
                if cont_shape.update(shape, frame):
                    found = True
                    del shapes[j]
                    break
            if not found:
                cont_shape.update_forced(frame)
            result[i] = cont_shape.get()
        for i in to_delete:
            del self.shapes[i]
        if self.track_new:
            for shape in shapes:
                self.shapes[self.__idx] = self.__create_continues_shape(shape, frame)
                result[self.__idx] = self.shapes[self.__idx].get()
                self.__idx += 1
        return result

    def get_shapes(self) -> Dict[int, Shape]:
        """
        returns the current location of the shapes based on continuity
        NOTE! this will be applied to the last frame given to the find_shapes method, only use this method if you need to get the shapes twice in an iteration

        :return: a dict mapping from unique ids to shapes
        """
        result = {}
        for i in self.shapes:
            result[i] = self.shapes[i].get()
        return result

    def get_shapes_as_list(self) -> List[Shape]:
        """
        gets all the shapes as a list instead of a dictionary
        :return: a list of all the tracked shapes (sorted by unique id's)
        """
        return list(self.get_shapes().values())



import abc
import time
from threading import Thread
from typing import Tuple

from gbvision.constants.types import Frame
from gbvision.utils.readable import Readable


class AsyncReadable(Readable, abc.ABC):
    """
    an async readable class that constantly reads on another thread
    """

    def __init__(self):
        self.__ok, self.__frame = False, None
        self.__thread = Thread(target=self.__async_read_wrapper)
        self.__thread.start()

    def read(self):
        return self.__ok, self.__frame

    def has_started_reading(self) -> bool:
        """
        checks if the async thread has started reading

        :return: True if the async thread started, False otherwise
        """
        return self.__ok

    def wait_start_reading(self, wait_time=0.01):
        """
        waits until the async thread starts reading

        :param wait_time: amounts of seconds to wait in every check interval, default is 0.01
        :return:
        """
        while not self.has_started_reading():
            time.sleep(wait_time)

    @abc.abstractmethod
    def _read(self) -> Tuple[bool, Frame]:
        """
        reads from the camera synchronously (similar to Readable.read), unsafe, not to use by the programmer
        this method will usually simply call super(self, ReadableClass).read()

        :return: tuple of bool (indicates if read was successful) and the frame (if successful, else None)
        """

    def __async_read_wrapper(self):
        while True:
            self.__ok, self.__frame = self._read()



import abc

from gbvision.constants.types import Frame
from .window import Window
from gbvision.models.system import EMPTY_PIPELINE


class RecordingWindow(Window, abc.ABC):

    def __init__(self, window_name: str, drawing_pipeline=EMPTY_PIPELINE, recording_pipeline=EMPTY_PIPELINE):
        Window.__init__(self, window_name=window_name, drawing_pipeline=drawing_pipeline)
        self.recording_pipeline = recording_pipeline

    @abc.abstractmethod
    def _record(self, frame: Frame):
        """
        unsafely records the given frame
        not to be used by the programmer
        :param frame: the frame to record
        """

    def show_frame(self, frame):
        self._record(self.recording_pipeline(frame))
        return Window.show_frame(self, frame)



from typing import Tuple, List, Union, Callable, TypeVar
from numpy import ndarray

Number = Union[int, float]
Point = Tuple[Number, Number]  # (x, y)
Contour = ndarray
Circle = Tuple[Tuple[Number, Number], Number]  # ((center_x, center_y), radius)
Rect = Tuple[Number, Number, Number, Number]  # (x, y, width, height)
Polygon = Union[List[Point], Contour]
FixedPolygon = List[Point]
RotatedRect = Tuple[Point, Point, float]  # ((center_x, center_y), (width, height), angle in degrees)
Ellipse = RotatedRect
Frame = Union[ndarray, None]
Color = Tuple[int, int, int]
FilterFunction = Callable[[Frame], Frame]
Coordinates = Tuple[int, int]
Location = Union[Tuple[Number, Number, Number], ndarray]
ROI = Tuple[int, int, int, int]
Shape = TypeVar('Shape')



import abc
from typing import Tuple

from gbvision.constants.types import Frame


class Readable(abc.ABC):
    """
    an interface representing an object you can read frames from, such as cameras and streams
    """

    @abc.abstractmethod
    def read(self) -> Tuple[bool, Frame]:
        """
        reads from the readable and returns the result

        :return: False, None if the operation was unsuccessful, else True, frame_result
        """



from typing import Union

import cv2
import numpy as np

from gbvision.constants.images import COLOR_TYPE
from gbvision.constants.types import Frame, Number, ROI
from gbvision.utils.thresholds import ColorThreshold
from gbvision.utils.thresholds.threshold import Threshold


def crop(frame: Frame, x: int, y: int, w: int, h: int) -> np.ndarray:
    """
    crops the image from (x, y) to (x+w, y+h)

    :param frame: the frame to crop
    :param x: the x coordinate to crop from
    :param y: the y coordinate to crop from
    :param w: the width of the cropped image
    :param h: the height of the cropped image
    :return: the cropped image
    """
    return frame[y:y + h, x:x + w]


def median_threshold(frame: Frame, stdv: Union[Number, np.ndarray],
                     box: Union[None, ROI] = None, color_encoding=ColorThreshold.THRESH_TYPE_BGR) -> Threshold:
    """
    finds a threshold using the median threshold method
    the median threshold method defines the lower bounds of the threshold as the median of a given region of the image
    minus some deviation variable, and the upper bounds as the same median plus the deviation variable
    in a mathematical term, the threshold is defined to be [median(X) - V, median(X) + V] where X is the frame region
    and V is the deviation variable

    :param frame: the frame
    :param stdv: the deviation variable, can be a scalar (same deviation for every channel) or a numpy array with the
        same size as the number of channels in the threshold, the deviation will be defined for each channel separately
    :param box: optional, a sub region of the frame from which the median is calculated, when set to None the median is
        calculated from the entire frame
    :param color_encoding: the type of color encoding the threshold should use, default is BGR
    :return: a Threshold object
    """
    if box is not None:
        frame = crop(frame, *box)
    color_encoding = color_encoding.upper()
    if color_encoding != ColorThreshold.THRESH_TYPE_BGR:
        frame = cv2.cvtColor(frame, COLOR_TYPE[color_encoding])
    med = np.median(frame, axis=(0, 1)).astype(int)
    if type(med) is not np.ndarray:
        med = np.array([med])
    params = list(map(lambda x: list(map(int, x)),
                      np.vectorize(lambda x: min(255, max(0, x)))(np.array([med - stdv, med + stdv])).T))
    return ColorThreshold(params, color_encoding)



from gbvision.utils.net import AsyncStreamReceiver
from .udp_stream_receiver import UDPStreamReceiver


class AsyncUDPStreamReceiver(AsyncStreamReceiver, UDPStreamReceiver):

    def __init__(self, port, *args, **kwargs):
        UDPStreamReceiver.__init__(self, port, *args, **kwargs)
        AsyncStreamReceiver.__init__(self, *args, **kwargs)

    def _read(self):
        return UDPStreamReceiver.read(self)



from typing import List, Callable

from gbvision.constants.types import RotatedRect, Polygon, Rect, Circle, Shape
from gbvision.utils.pipeline import PipeLine
import numpy as np
import cv2


class __InnerShapeFilter(PipeLine):
    """
    filters out all the shapes that are colliding with shapes with a lower (smaller) index
    maps from List[Shape] to List[Shape], where the output list is the input list without the colliding shapes
    usually used on a sorted list, to remove any shape that is inside another shape

    """

    def __init__(self, collision_func: Callable[[Shape, Shape], bool]):
        def _filter(shapes: List[Shape]) -> List[Shape]:
            filtered_shapes = []
            for i, shape in enumerate(shapes):
                shape_invalid = False
                for j in range(i):
                    shape_invalid = collision_func(shape, shapes[j])
                    if shape_invalid:
                        break
                if not shape_invalid:
                    filtered_shapes.append(shape)
            return filtered_shapes

        PipeLine.__init__(self, _filter)


def circle_collision(circ1: Circle, circ2: Circle) -> bool:
    """
    detects if two circles are colliding

    :param circ1: the first circle
    :param circ2: the second circle
    :return: True if circles are colliding, False otherwise
    """
    center1, r1 = circ1
    center2, r2 = circ2
    return (center1[0] - center2[0]) ** 2 + (center1[1] - center2[1]) ** 2 < (r1 + r2) ** 2


filter_inner_circles = __InnerShapeFilter(circle_collision)


def rect_collision(r1: Rect, r2: Rect) -> bool:
    """
    detects if two rects are colliding

    :param r1: the first rect
    :param r2: the second rect
    :return: True if the rects are colliding, False otherwise
    """
    return not (r1[0] > r2[0] + r2[2] or
                r1[0] + r1[2] < r1[0] or
                r1[1] > r2[1] + r2[3] or
                r1[1] + r1[3] < r2[1])


filter_inner_rects = __InnerShapeFilter(rect_collision)


def convex_shape_collision(shape1: Polygon, shape2: Polygon) -> bool:
    """
    detects collision between two convex shapes
    Note: if you are uncertain if a shape is convex, use convex_hull on it, it will make it convex

    :param shape1: the first shape, as a contour
    :param shape2: the second shape, as a contour
    :return: True if the shapes are colliding, False otherwise
    """
    shape1, shape2 = np.array(shape1), np.array(shape2)
    shape1 = shape1.reshape(shape1.size // 2, 2)
    shape2 = shape2.reshape(shape1.size // 2, 2)

    projection_axises = []

    negation_array = np.array([-1.0, 1.0])

    for i in range(len(shape1)):
        projection_axises.append(shape1[(i + 1) % len(shape1)] - shape1[i])
    for i in range(len(shape2)):
        projection_axises.append(shape2[(i + 1) % len(shape2)] - shape2[i])

    for i in range(len(projection_axises)):
        projection_axises[i] = projection_axises[i][::-1] * negation_array

    for axis in projection_axises:
        proj1 = []
        for vertix in shape1:
            proj1.append(vertix.dot(axis))
        proj1 = min(proj1), max(proj1)

        proj2 = []
        for vertix in shape2:
            proj2.append(vertix.dot(axis))
        proj2 = min(proj2), max(proj2)

        if proj1[0] > proj2[1] or proj2[0] > proj1[1]:
            return False

    return True


filter_inner_convex_shapes = __InnerShapeFilter(convex_shape_collision)


def rotated_rect_collision(rr1: RotatedRect, rr2: RotatedRect) -> bool:
    """
    detects if two rotated rects are colliding

    :param rr1: the first rotated rect
    :param rr2: the second rotated rect
    :return: True if the shapes are colliding, False otherwise
    """
    return convex_shape_collision(cv2.boxPoints(rr1), cv2.boxPoints(rr2))


filter_inner_rotated_rects = __InnerShapeFilter(rotated_rect_collision)



from .circle_finder import CircleFinder
from .contour_finder import ContourFinder
from .convex_polygon_finder import ConvexPolygonFinder
from .object_finder import ObjectFinder
from .polygon_finder import PolygonFinder
from .rect_finder import RectFinder
from .rotated_rect_finder import RotatedRectFinder




import abc

from gbvision.constants.types import Rect, Number, Frame, Point, Shape
from gbvision.utils.shapes.base_shape import BaseShapeType
from gbvision.utils.tracker import Tracker


class ContinuesShape(abc.ABC):
    """
    An abstract class who's target is to determine in which case two shapes are describing
    the same object, and by that represent a shape in a continues matter.

    in case that no shape matching the object is found then the tracker is deployed by using force_update.

    :param shape: the shape to track with continuity
    :param frame: the frame from which the shape was taken
    :param tracker: optional, a tracker used to track the object if it wasn't found with continuity (default is an empty tracker)
    :param max_area_ratio: the maximum ration between areas of the current shape and another shape \
     so that they can be thought of as the same, default is 2.0
    :param max_distance_ratio: the maximum ratio between the sum of areas of this shape and another and their distance, default is 0.1
    """

    def __init__(self, shape, frame: Frame, tracker: Tracker = None, max_area_ratio=2.0,
                 max_distance_ratio=2.0):  # initialization method of the abstract class
        assert max_area_ratio > 1.0  # sets maximum area ratio as 1
        self._shape = shape  # shape describing the object
        self._count = 0  # the count of how many frames the object cannot be found
        self._tracker = Tracker() if tracker is None else tracker  # Tracker setup in case object is not found
        self._tracker.init(frame, self._to_bounding_rect(shape))
        self.max_area_ratio = max_area_ratio  # setting the maximum bound to which the area difference between the
        # two shapes can be withstanded until it declared as not the same object
        self.max_distance_ratio = max_distance_ratio  # same goes here except it refers to distance between shapes

    @staticmethod
    @abc.abstractmethod
    def _base_shape() -> BaseShapeType:
        """
        returns the base shape matching this continues shape

        :return: the base shape (a class that inherits from BaseShape)
        """

    def _shape_collision(self, shape: Shape) -> bool:
        return self._base_shape().shape_collision(self._shape, shape)

    @classmethod
    def _shape_area(cls, shape: Shape) -> Number:
        return cls._base_shape().shape_area(shape)

    @classmethod
    def _shape_center(cls, shape: Shape) -> Point:
        return cls._base_shape().shape_center(shape)

    @staticmethod
    @abc.abstractmethod
    def _from_bounding_rect(bounding_rect: Rect) -> Shape:
        """
        a function which finds a shape according to its bounding rectangle
        :param: the rectangle bounding the shape
        :return: a shape that is bound by the bounding rect
        """

    @staticmethod
    @abc.abstractmethod
    def _to_bounding_rect(shape: Shape) -> Rect:
        """
        finds a rectangle bounding a shape
        :param shape: the shape's bounding rectangle you wish to find
        :return: a rectangle which bounds the shape given
        """

    def _shape_square_distance(self, other_shape: Shape) -> Number:
        """
        a method which finds the distance between the centers of the object shape and another one
        :param other_shape: the other shape you want to check the distance to
        :return: the distance between the shape squared in order to save computing of square root (pythagorean theorem)
        """
        self_center, other_center = self._shape_center(self._shape), self._shape_center(other_shape)
        return (other_center[0] - self_center[0]) ** 2 + \
               (other_center[1] - self_center[1]) ** 2

    def _is_legal(self, shape: Shape) -> bool:
        """
        checks a variety of different relations between to shapes to determine whether it describes the same object or not
        :param shape: the other shape comparing to the current one
        :return: true or false, same or different
        """
        if self._shape_collision(shape):
            if 1.0 / self.max_area_ratio <= self._shape_area(self._shape) / self._shape_area(
                    shape) <= self.max_area_ratio:
                if self._shape_square_distance(shape) <= (
                        self._shape_area(self._shape) + self._shape_area(shape)) * self.max_distance_ratio:
                    return True
        return False

    def get(self) -> Shape:
        """
        retrieve the shape that this continues shape tracks

        :return: this continues shape's inner shape
        """
        return self._shape

    def update(self, shape: Shape, frame: Frame) -> bool:
        """
        an annual check updating the location and data of the object
        :param shape: the shape suspect as the same object
        :param frame: the frame on which the suspect shape is
        :return: true or false, same shape, not the same shape
        """
        if self._is_legal(shape):
            self._shape = shape
            self._count = 0
            self._tracker.init(frame, self._to_bounding_rect(shape))
            return True
        return False

    def update_forced(self, frame: Frame):
        """
        an update which happens when you lost the shape with continuity
        :param frame: the frame on which opencv2 tracking is happening
        """
        self._shape = self._from_bounding_rect(self._tracker.update(frame))
        self._count += 1

    def is_lost(self, max_count: int) -> bool:
        """
        check if it has been too long (more that max_count frames) since you last saw the object

        :param max_count:the maximum amount of frames tolerable before the shape is declared lost, None is infinity
        :return: lost or not
        """
        return max_count is not None and self._count > max_count



from .async_camera import AsyncCamera
from .async_usb_camera import AsyncUSBCamera
from .camera import Camera
from .camera_data import CameraData
from .camera_list import CameraList
from .empty_camera import EmptyCamera
from .stream_camera import StreamCamera, SimpleStreamCamera
from .usb_camera import USBCamera
from .usb_stream_camera import USBStreamCamera



from .wrapper_opencv_window import WrapperOpenCVWindow
from gbvision.utils.readable import Readable


class ReadableWindow(WrapperOpenCVWindow):
    """
    a basic window that displays the stream from any readable

    :type self.wrap_object: gbvision.Readable
    """

    def __init__(self, window_name: str, wrap_object: Readable, *args, **kwargs):
        WrapperOpenCVWindow.__init__(self, window_name, wrap_object, *args, **kwargs)

    def _get_frame(self):
        _, frame = self.wrap_object.read()
        return frame



from typing import Callable, Iterable, T


def split_list(f: Callable[[T], int], lst: Iterable[T], amount=2):
    """
    splits the list into several list according to the function f

    :param lst: the list to split
    :param f: a function which maps from an argument to the index of the list it should go to
        for example if we wanted a function to split a list into a list of positive and negative number f could look like
        lambda x: int(x >= 0)
    :param amount: the amount of lists to split the data to (2 by default)
    :return: a tuple of all the lists created,
    """
    temp = tuple([] for _ in range(amount))
    for i in lst:
        temp[f(i)].append(i)
    return temp



import abc
import pickle
import struct
import cv2
import time

from gbvision.constants.math import EPSILON
from gbvision.constants.types import Frame


class StreamBroadcaster(abc.ABC):
    """
    this is an abstract broadcaster that sends stream to a broadcast receiver
    this class should not be instanced but inherited from
    creates a new stream broadcaster with all parameters that are used in every broadcaster
    
    :param shape: optional, the shape (x, y) of the sent frame, when set to something other then (0, 0) it overrides
        the fx and fy parameters, when set to (0, 0) it is not used
    :param fx: ratio between width of the given frame to the width of the frame sent, default is 1 (same width)
    :param fy: ratio between height of the given frame to the height of the frame sent, default is 1 (same height)
    :param use_grayscale: boolean indicating if the frame should be converted to grayscale when sent,
        default is False
    :param max_fps: integer representing the maximum fps (frames per second) of the stream, when set to None
        there is no fps limitation, default is None
    :param max_bitrate: Integer that determines the max bitrate of video stream.
        The bitrate is messured with Kbps and default is None.
    """

    def __init__(self, shape=(0, 0), fx: float = 1.0, fy: float = 1.0, use_grayscale: bool = False,
                 max_fps: int = None, im_encode='.jpg', max_bitrate: int = None):
        self.shape = shape
        self.fx = fx
        self.fy = fy
        self.use_grayscale = use_grayscale
        self.max_fps = max_fps
        self.prev_time = 0.0
        self.im_encode = im_encode
        self.max_bitrate = max_bitrate

    def send_frame(self, frame: Frame):
        if frame is not None:
            frame = self._prep_frame(frame)
            frame = cv2.imencode(self.im_encode, frame)[1]
        data = pickle.dumps(frame)
        data = struct.pack("I", len(data)) + data
        if self._can_send_frame(data):
            self._send_frame(data)
            self._update_time()

    @abc.abstractmethod
    def _send_frame(self, frame: bytes):
        """
        unsafely sends the given frame (as pickled data) to the stream receiver
        should not be used by the programmer, only by the api

        :param frame: the frame to send as pickled data
        """
        pass

    def _prep_frame(self, frame):
        """
        prepares an image to be sent
        resize and convert the colors of the image by the parameters of the stream broadcaster

        :param frame: the frame to prepare
        :return: the frame after preparation
        """
        frame = cv2.resize(frame, self.shape, fx=self.fx, fy=self.fy)
        if self.use_grayscale and len(frame.shape) > 2:
            frame = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
        return frame

    def _legal_time(self) -> bool:
        """
        checks if at the fps will not pass the max fps limit if an image will be sent at the current moment
        
        :return: true if the image can be sent, false otherwise
        """
        return self.max_fps is None or (time.time() - self.prev_time) * self.max_fps >= 1

    def _update_time(self):
        """
        updates the previous time a frame was sent, used at the end of send_frame
        """
        self.prev_time = time.time()

    def _legal_bitrate(self, frame: bytes):
        """
        :return: True if there's no bitrate limit or frame bitrate is below max bitrate.  
        """
        return self.max_bitrate is None or len(frame) / (
                (time.time() - self.prev_time + EPSILON) * 1000) <= self.max_bitrate

    def _can_send_frame(self, frame: bytes):
        if not self._legal_time():
            return False

        if not self._legal_bitrate(frame):
            return False

        return True



import socket

from gbvision.constants.net import LOCAL_SERVER_IP
from .stream_receiver import StreamReceiver


class UDPStreamReceiver(StreamReceiver):
    """
    this class uses UDP to receive a stream over the network, the stream is by default set to be MJPEG
    the broadcaster is the client and the receiver is the server
    WARNING: do not use this class to send large images, udp has a limited packet size

    :param port: the port which udp should use
    """

    def __init__(self, port: int, *args, **kwargs):
        StreamReceiver.__init__(self, *args, **kwargs)
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.server_addr = (LOCAL_SERVER_IP, port)
        self.socket.bind(self.server_addr)

    def _receive(self) -> bytes:
        return self.socket.recv(2 ** 20)



import abc

from gbvision.utils.net.stream_receiver import StreamReceiver
from gbvision.utils.async_readable import AsyncReadable


class AsyncStreamReceiver(AsyncReadable, StreamReceiver, abc.ABC):
    """
    an abstract async tcp stream receiver that receives frames on another thread
    None! when inheriting from this class and another StreamReceiver class, make sure you call the other class'
    constructor before this class' constructor, but also make sure you inherit from this class first in order
    """

    def __init__(self, *args, **kwargs):
        StreamReceiver.__init__(self, *args, **kwargs)
        AsyncReadable.__init__(self)



from typing import Union, Tuple
import cv2
import numpy as np

from gbvision.utils.pipeline import PipeLine


class Erode(PipeLine):
    """
    creates a pipeline that erodes the image by a kernel of ones
    used mainly for Erode & Dilate denoise filters

    :param ksize: the kernel size, either an integer (meaning an nxn kernel) or a tuple (nxm kernel)
    :param iterations: optional, the amount of Erode iterations to perform, default is 1
        None! a large number of iterations will slow down the program
    :return: a pipeline that erodes the given frame
    """

    def __init__(self, ksize: Union[int, Tuple[int, int]], iterations=1):
        if type(ksize) is int:
            ksize = (ksize, ksize)
        PipeLine.__init__(self, lambda frame: cv2.erode(frame, np.ones(ksize), iterations=iterations))


class Dilate(PipeLine):
    """
        creates a pipeline that dilates the image by a kernel of ones
        used mainly for Erode & Dilate denoise filters

        :param ksize: the kernel size, either an integer (meaning an nxn kernel) or a tuple (nxm kernel)
        :param iterations: optional, the amount of Dilate iterations to perform, default is 1
            None! a large number of iterations will slow down the program
        :return: a pipeline that dilates the given frame
        """

    def __init__(self, ksize: Union[int, Tuple[int, int]], iterations=1):
        if type(ksize) is int:
            ksize = (ksize, ksize)
        PipeLine.__init__(self, lambda frame: cv2.dilate(frame, np.ones(ksize), iterations=iterations))


class MedianBlur(PipeLine):
    """
    creates a pipeline that blurs the given frame using the median blur method
    works very good for denoising purposes

    :param ksize: the size of the kernel used by the filter, must be an odd number
    :return: a pipeline that filters the image using the median blur method
    """

    def __init__(self, ksize: int):
        PipeLine.__init__(self, lambda frame: cv2.medianBlur(frame, ksize))



import numpy as np

from gbvision.models.shapes import circle_collision
from gbvision.constants.types import Rect, Circle
from .continues_shape import ContinuesShape
from gbvision.utils.shapes.base_circle import BaseCircle


class ContinuesCircle(ContinuesShape):
    """
    An implementation of ContinuesShape to circles.
    used to try and check whether two circles are indeed the same one.
    """

    @staticmethod
    def _base_shape():
        return BaseCircle

    def __init__(self, shape: Circle, *args, **kwargs):
        ContinuesShape.__init__(self, shape, *args, **kwargs)

    def _shape_collision(self, shape) -> bool:
        return circle_collision(self._shape, shape)

    @staticmethod
    def _from_bounding_rect(bounding_rect) -> Circle:
        circle = (
            (bounding_rect[0] + bounding_rect[2] / 2, bounding_rect[1] + bounding_rect[3] / 2),
            (bounding_rect[3] + bounding_rect[2]) / 4)
        return circle

    @staticmethod
    def _to_bounding_rect(shape) -> Rect:
        rect = (shape[0][0] - shape[1], shape[0][1] - shape[1], 2 * shape[1], 2 * shape[1])
        return rect



import abc
from typing import Tuple

from gbvision.constants.types import Frame
from gbvision.utils.net.stream_broadcaster import StreamBroadcaster
from .camera import Camera


class StreamCamera(Camera, abc.ABC):
    """
    an abstract class that represents a streaming camera
    the streaming camera is very similar to a regular camera, but has an option
    which allows it to stream the frames when it reads them
    """

    @abc.abstractmethod
    def is_streaming(self) -> bool:
        """
        checks if the camera is currently streaming

        :return: True if camera is streaming, otherwise False
        """

    @abc.abstractmethod
    def toggle_stream(self, should_stream: bool):
        """
        turn on or off the stream feature

        :param should_stream: True to activate stream, False to deactivate
        """

    def read(self):
        ok, frame = self._read()
        self._stream(frame)
        return ok, frame

    @abc.abstractmethod
    def _read(self) -> Tuple[bool, Frame]:
        """
        unsafely reads from the camera, not to use by the programmer, only by the api
        usually this function is a set to return the value of super(self, CameraClass).read()

        :return: the return value of Camera.read: (ok, frame)
        """

    @abc.abstractmethod
    def _stream(self, frame: Frame):
        """
        unsafely streams a frame, not to use by the programmer, only by the api

        :param frame: the frame to stream
        """


class SimpleStreamCamera(StreamCamera, abc.ABC):
    """
    a simple abstract camera that uses a gbvision.StreamBroadcaster to send streams
    this class is abstract and cannot exist on it's own, you must inherit from it and implement the _read method

    for example:

    Example::
        class USBStreamCamera(SimpleStreamCamera, USBCamera):
            def _read(self) -> Tuple[bool, Frame]:
                return USBCamera.read(self)

            def __init__(self, broadcaster, port, data=UNKNOWN_CAMERA):
                SimpleStreamCamera.__init__(self, broadcaster)
                USBCamera.__init__(port, data)
    """

    def __init__(self, broadcaster: StreamBroadcaster, should_stream=False):
        self.__is_streaming = should_stream
        self.stream_broadcaster = broadcaster

    def is_streaming(self):
        return self.__is_streaming

    def toggle_stream(self, should_stream: bool):
        self.__is_streaming = should_stream

    def _stream(self, frame):
        self.stream_broadcaster.send_frame(frame)



from abc import ABC

from .recording_opencv_window import RecordingOpenCVWindow
from .wrapper_opencv_window import WrapperOpenCVWindow
from gbvision.models.system import EMPTY_PIPELINE


class RecordingWrapperOpenCVWindow(RecordingOpenCVWindow, WrapperOpenCVWindow, ABC):
    def __init__(self, window_name: str, wrap_object, file_name: str, fps=20.0, exit_button='qQ',
                 drawing_pipeline=EMPTY_PIPELINE, recording_pipeline=EMPTY_PIPELINE, width=0, height=0):
        RecordingOpenCVWindow.__init__(self, window_name, file_name=file_name, fps=fps,
                                       recording_pipeline=recording_pipeline, width=width, height=height)
        WrapperOpenCVWindow.__init__(self, window_name=window_name, wrap_object=wrap_object, exit_button=exit_button,
                                     drawing_pipeline=drawing_pipeline)



from .async_stream_receiver import AsyncStreamReceiver
from .async_tcp_stream_receiver import AsyncTCPStreamReceiver
from .async_udp_stream_receiver import AsyncUDPStreamReceiver
from .stream_broadcaster import StreamBroadcaster
from .stream_receiver import StreamReceiver
from .tcp_stream_broadcaster import TCPStreamBroadcaster
from .tcp_stream_receiver import TCPStreamReceiver
from .udp_stream_broadcaster import UDPStreamBroadcaster
from .udp_stream_receiver import UDPStreamReceiver



from abc import ABC

from .window import Window
import cv2
from gbvision.models.system import EMPTY_PIPELINE


class OpenCVWindow(Window, ABC):

    def __init__(self, window_name: str, exit_button='qQ', drawing_pipeline=EMPTY_PIPELINE, flags=cv2.WINDOW_FREERATIO):
        Window.__init__(self, window_name=window_name, drawing_pipeline=drawing_pipeline)
        self.exit_button = exit_button
        self.flags = flags
        self.last_key_pressed = None

    def _show_frame(self, frame):
        if frame is None:
            return False
        cv2.imshow(self.window_name, frame)
        k = chr(cv2.waitKey(1) & 0xFF)
        self.last_key_pressed = k
        if k in self.exit_button:
            return False
        return True

    def _open(self):
        cv2.namedWindow(self.window_name, self.flags)

    def _close(self):
        cv2.destroyWindow(self.window_name)



import abc
from .camera import Camera
from gbvision.utils.async_readable import AsyncReadable


class AsyncCamera(Camera, AsyncReadable, abc.ABC):
    """
    an abstract class that represents an async camera
    the async camera is similar to a regular camera, but executes the read actions on another thread
    thus not blocking the processing thread

    you will usually want to inherit from this class and from another camera class in order to use the async functionality
    for example:

    Example::
        class AsyncUSBCamera(AsyncCamera, USBCamera):
            def _read(self) -> Tuple[bool, Frame]:
                return USBCamera.read(self)

            def __init__(self, port, data=UNKNOWN_CAMERA):
                USBCamera.__init__(self, port, data)
                AsyncCamera.__init__(self)
    """

    def __init__(self):
        AsyncReadable.__init__(self)



from .recording_window import RecordingWindow
from .opencv_window import OpenCVWindow
from gbvision.models.system import EMPTY_PIPELINE
from os.path import splitext
import cv2
from gbvision.constants.video import VIDEO_FILE_TYPE


class RecordingOpenCVWindow(OpenCVWindow, RecordingWindow):
    """
    :param file_name: the name of the output file
    :param drawing_pipeline: optional, a pipeline of drawing functions that will run on the frame before displaying
        it
    :param exit_button: an array of keys (a string), when one of the keys are pressed the window will be closed
    :param window_name: the title of the window
    :param fps: the fps of the video file
    :param recording_pipeline: optional, a drawing pipeline to run on the frames being recorded
    """

    def __init__(self, window_name: str, file_name: str, fps=20.0, exit_button='qQ',
                 drawing_pipeline=EMPTY_PIPELINE, recording_pipeline=EMPTY_PIPELINE, width=0, height=0,
                 flags=cv2.WINDOW_FREERATIO):
        """
        initializes the stream window
        
        """
        OpenCVWindow.__init__(self, window_name, exit_button=exit_button, flags=flags)
        RecordingWindow.__init__(self, window_name=window_name, drawing_pipeline=drawing_pipeline,
                                 recording_pipeline=recording_pipeline)
        self.file_name = file_name

        _, file_ext = splitext(file_name)

        self.fourcc = cv2.VideoWriter_fourcc(*VIDEO_FILE_TYPE[file_ext.upper()])

        self.fps = fps

        self.video_writer = None
        self.width = width
        self.height = height

    def _record(self, frame):
        if frame is None or len(frame.shape) == 0:
            return
        if self.video_writer is None:
            width = self.width if self.width != 0 else frame.shape[1]
            height = self.height if self.height != 0 else frame.shape[0]
            self.video_writer = cv2.VideoWriter(self.file_name, self.fourcc, self.fps, (width, height))
        self.video_writer.write(frame)



from .recording_readable_window import RecordingReadableWindow
from .camera_window import CameraWindow


class RecordingCameraWindow(RecordingReadableWindow, CameraWindow):
    """
    a basic window that displays the video from a camera
    and records the video to a file
    """



import cv2
import numpy as np

from gbvision.constants.math import EPSILON
import gbvision.utils.cameras as cameras
from gbvision.models.contours import contour_center
from gbvision.constants.types import Point, Contour, Number, Location


class GameObject:
    """
    constructor of the image object
    which is an object on field

    :param area: the square root of the surface area of the object in real life

    """

    def __init__(self, area: Number):
        self.area = area

    def distance_by_contours(self, camera: cameras.Camera, cnt: Contour) -> float:
        """
        Note: this measures the distance between the camera and the object, to use another measuring point
        calculate the norm of the location
        
        :param camera: the camera, can be either Camera or CameraList
        :param cnt: the contours of this object in the frame
        :return: the norm of the vector between the camera and the object (in meters)
        """
        return self.distance_by_params(camera, np.sqrt(cv2.contourArea(cnt)))

    def distance_by_params(self, camera: cameras.Camera, area: Number) -> float:
        """
        Note: this measures the distance between the camera and the object, to use another measuring point
        calculate the norm of the location

        :param camera: the camera, can be either Camera or CameraList
        :param area: a float representing the square root of the area of the object
            (in pixels)
        :return: the norm of the vector between the camera and the object (in meters)
        """
        return camera.get_data().focal_length * self.area / (area + EPSILON)

    def location_by_contours(self, camera: cameras.Camera, cnt: Contour) -> Location:
        """
        :param camera: the camera, can be either Camera or CameraList
        :param cnt: the contours of this object in the frame
        :return: a 3d vector of the relative [x y z] location between the object and the camera (in meters)
        """
        return self.location_by_params(camera, np.sqrt(cv2.contourArea(cnt)), contour_center(cnt))

    def location_by_params(self, camera: cameras.Camera, area: Number, center: Point) -> Location:
        """
        :param camera: the camera, can be either Camera or CameraList
        :param area: a float representing the square root of the area of the object \
            (in pixels)
        :param center: the center (x,y) of this object in the frame
        :return: a 3d vector of the relative [x y z] location between the object and the camera/measuring point (in meters)
        """
        frame_center = camera.get_width(), camera.get_height()
        frame_center = np.array(frame_center) / 2
        x, y = np.array(center) - frame_center
        alpha = x * camera.get_data().fov / frame_center[0]
        beta = y * camera.get_data().fov / frame_center[1]
        rel = np.array([[np.sin(alpha), np.sin(beta),
                         np.sqrt(1 - np.sin(alpha) ** 2 - np.sin(beta) ** 2)]]) * self.distance_by_params(camera, area)
        return camera.get_data().rotation_matrix.dot(rel.T).flatten() + camera.get_data().offset



import abc
from typing import List, Callable

import cv2

from gbvision.constants.types import FilterFunction, Frame


class Threshold(abc.ABC):
    """
    a class that represents a function that maps from an image to a binary image
    where 255 states that the pixel at the original image was in a range represented by the threshold object
    and 0 states the pixel was out of the range
    for example:
    ColorThreshold([[200, 255], [0, 50], [0, 50]], 'RGB')
    the above threshold represents a relatively red pixel. when an image is filtered by it, every pixel that is
    relatively red will be given the value 255, and every pixel that isn't will be given the value of 0
    """

    @abc.abstractmethod
    def __call__(self, frame: Frame) -> Frame:
        """
        activates the threshold filter on the given image
        :param frame: the image to activate the threshold on
        :return: a binary image, the frame after the threshold filter
        """

    def __or__(self, other: 'Threshold') -> 'Threshold':
        return ThresholdGroup(cv2.bitwise_or, self, other)

    def __and__(self, other: 'Threshold') -> 'Threshold':
        return ThresholdGroup(cv2.bitwise_and, self, other)


class ThresholdGroup(Threshold):
    """
    a class that constructs a threshold filter out of several threshold filters and a binary mask function
    for example, use of the function on two thresholds with the binary function "bitwise_or" will result
    in a filter that outputs 255 for a pixel if it is in either one of the threshold's range
    using the "bitwise_and" function will output 255 for a pixel only if it is in both the threshold's range

    :param thresholds: all the thresholds to join in the threshold group
    :param binary_mask: a binary function that maps from a pair of binary images to a single binary image
        default value is cv2.bitwise_or
    """

    def __init__(self, binary_mask: Callable[[Frame, Frame], Frame], *thresholds):
        self.binary_mask = binary_mask
        self.thresholds: List[Threshold] = list(thresholds)

    def __call__(self, frame: Frame) -> Frame:
        """
        apply the threshold filter to the frame
        :param frame: the frame to apply the filter to
        :return: the binary image, the frame after the threshold group filter
        """
        if len(self.thresholds) == 0:
            return frame
        frame_tag = self.thresholds[0](frame)
        for i in range(1, len(self.thresholds)):
            frame_tag = self.binary_mask(frame_tag, self.thresholds[i](frame))
        return frame_tag

    def __iter__(self):
        """
        :return: an iterator that iterates through all this group's filters
        """
        return iter(self.thresholds)

    def __len__(self):
        """
        :return: the amount of filters in this threshold group
        """
        return len(self.thresholds)

    def __getitem__(self, item: int) -> FilterFunction:
        return self.thresholds[item]

    def __setitem__(self, key: int, value: FilterFunction):
        self.thresholds[key] = value



from .basic_ops import gray, sharpen, blue, blur, corners, red, edges, green
from .cameras import LIFECAM_STUDIO, LIFECAM_3000, UNKNOWN_CAMERA
from .contours import contours_to_rotated_rects, contours_to_rotated_rects_sorted, fix_contours_shape, \
    sort_rotated_rects, contour_center, contours_centers, contours_to_circles, contours_to_circles_sorted, \
    contours_to_ellipses, contours_to_ellipses_sorted, contours_to_polygons, contours_to_rects, \
    contours_to_rects_sorted, convex_hull, convex_hull_multiple, sort_circles, find_contours, polygon_center, \
    polygons_centers, sort_contours, sort_rects, sort_ellipses, sort_polygons, FilterContours
from .denoising import Erode, Dilate, MedianBlur
from .shapes import convex_shape_collision, filter_inner_convex_shapes, filter_inner_rotated_rects, \
    rotated_rect_collision, circle_collision, filter_inner_circles, filter_inner_rects, rect_collision
from .system import EMPTY_PIPELINE



import cv2


class __EmptyTracker:
    def __init__(self):
        self.__rect = None

    def init(self, frame, rect):
        self.__rect = rect
        return True

    def update(self, frame):
        return True, self.__rect


TRACKER_ALGORITHMS = {
    "BOOSTING": cv2.TrackerBoosting_create,
    "MIL": cv2.TrackerMIL_create,
    "KCF": cv2.TrackerKCF_create,
    "TLD": cv2.TrackerTLD_create,
    "MEDIANFLOW": cv2.TrackerMedianFlow_create,
    "GOTURN": cv2.TrackerGOTURN_create,
    "MOSSE": cv2.TrackerMOSSE_create,
    "CSRT": cv2.TrackerCSRT_create,
    "EMPTY": __EmptyTracker
}


class Tracker:
    """
    a tracker object that tracks a rectangle in a video using an opencv tracking algorithm

    :param tracker_type: Tracker algorithm taken from this list: BOOSTING, MIL, KCF, TLD, MEDIANFLOW,
        GOTURN, MOSSE, CSRT, EMPTY. (Default is EMPTY)
    """

    TRACKER_TYPE_BOOSTING = 'BOOSTING'
    TRACKER_TYPE_MIL = 'MIL'
    TRACKER_TYPE_KCF = 'KCF'
    TRACKER_TYPE_TLD = 'TLD'
    TRACKER_TYPE_MEDIANFLOW = 'MEDIANFLOW'
    TRACKER_TYPE_GOTURN = 'GOTURN'
    TRACKER_TYPE_MOSSE = 'MOSSE'
    TRACKER_TYPE_CSRT = 'CSRT'
    TRACKER_TYPE_EMPTY = 'EMPTY'

    def __init__(self, tracker_type="EMPTY"):
        tracker_type = tracker_type.upper()
        assert tracker_type in TRACKER_ALGORITHMS
        self.tracker = TRACKER_ALGORITHMS[tracker_type]()
        self.tracker_type = tracker_type

    def init(self, frame, rect):
        """
        Initlize the tracker

        :param frame: The frame
        :param rect: Given rectangle
        :return: True if initialization went succesfully, false otherwise
        """
        return self.tracker.init(frame, tuple([int(max(x, 0)) for x in rect]))

    def update(self, frame):
        """
        Get the rect location in new frame

        :param frame: the frame
        :return: The location of the rect in new frame

        """
        return self.tracker.update(frame)[1]



import abc

from gbvision.models.system import EMPTY_PIPELINE
from gbvision.constants.types import Frame


class Window(abc.ABC):
    """
    an abstract window class

    :param window_name: the title of the new window
    :param drawing_pipeline: optional, a pipeline that draws on each frame before displaying it
    """

    def __init__(self, window_name: str, drawing_pipeline=EMPTY_PIPELINE):
        """
        initializes the window
        """
        self.window_name = window_name
        self._is_opened = False
        self.drawing_pipeline = drawing_pipeline

    @abc.abstractmethod
    def _show_frame(self, frame: Frame) -> bool:
        """
        shows the frame
        :param frame: the frame to show
        :return: False if the window should be closed, True otherwise
        """
        pass

    def is_opened(self) -> bool:
        """

        :return: True is the window is opened, False otherwise
        """
        return self._is_opened

    def show_frame(self, frame: Frame) -> bool:
        """
        shows the frame on the window
        :param frame: the frame to show
        :return: false if the window was closed, true otherwise
        """
        if not self.is_opened():
            self.open()
        if self._show_frame(self.drawing_pipeline(frame)):
            return True
        self.close()
        return False

    @abc.abstractmethod
    def _open(self):
        """
        unsafely opens the window
        not to be used by the programmer, only by the function open
        """
        pass

    @abc.abstractmethod
    def _close(self):
        """
        unsafely closes the window
        not to be used by the programmer, only by the function close
        """
        pass

    def open(self):
        """
        opens the window
        """
        self._open()
        self._is_opened = True

    def close(self):
        """
        closes the window
        """
        self._close()
        self._is_opened = False


    def __del__(self):
        self.close()


from typing import Tuple

from gbvision.constants.types import Frame
from .usb_camera import USBCamera
from .async_camera import AsyncCamera
from gbvision.models.cameras import UNKNOWN_CAMERA


class AsyncUSBCamera(AsyncCamera, USBCamera):
    """
    A simple async usb camera
    """

    def _read(self) -> Tuple[bool, Frame]:
        return USBCamera.read(self)

    def __init__(self, port, data=UNKNOWN_CAMERA):
        USBCamera.__init__(self, port, data)
        AsyncCamera.__init__(self)



import numpy as np

from gbvision.constants.types import Location


def distance_from_object(loc: Location) -> float:
    """
    the absolute distance from the camera to this object

    :param loc: the object's location (2d or 3d)
    :returns: the absolute distance (float)
    """
    if len(loc) <= 3:
        return np.linalg.norm(loc)
    return np.linalg.norm(loc[:3])


def plane_angle_by_location(loc: Location) -> float:
    """
    calculates the angle from the camera to the object's projection on the x-z plane (y=0 plane)

    :param loc: the 3d location
    :return: the angle (in radians)
    """
    return np.arctan(loc[0] / loc[2])


def plane_distance_from_object(loc: Location) -> float:
    """
    calculates the distance from the object's projection on the x-z plane (y=0 plane)
    the distance on the y axis is ignored in this calculation
    
    :param loc: the 3d location
    :return: the distance without regarding the y axis
    """
    return np.sqrt(loc[0] ** 2 + loc[2] ** 2)


def viewing_angle_of_object(part1: Location, part2: Location, x_distance: float) -> float:
    """
    finds the viewing angle of an object based on a split of the object to two parts (part1 and part2), and the x \
    distance between those two parts in real life (in meters)

    :param part1: the first part of the object (the left one)
    :param part2: the second part of the object (the right one)
    :param x_distance: the distance between the two objects in meters

    :return: the viewing angle of the object
    """
    return np.pi / 2 - np.arccos(max(-1, min(1, (part1[2] - part2[2]) / x_distance)))



from typing import Tuple, Union

from .camera import Camera
from .camera_data import CameraData
from gbvision.constants.types import Frame


class EmptyCamera(Camera):
    """
    a camera class used for testing, it cannot be read from but can be used for location finding with finders and game\
     objects, also used for measuring of distances when using streams instead of cameras to read frames

    :param data: the camera's CameraData object, should match the fake camera's data
    :param width: the width of a frame read from the fake camera
    :param height: the height of a frame read from the fake camera
    """

    def __init__(self, data: CameraData, width: int, height: int):
        self.data = data.copy()
        self.width = width
        self.height = height

    def release(self):
        pass

    def is_opened(self) -> bool:
        return False

    def set_exposure(self, exposure: Union[int, float, bool]) -> bool:
        return False

    def set_auto_exposure(self, auto: Union[int, float, bool]) -> bool:
        return False

    def get_data(self) -> CameraData:
        return self.data

    def get_width(self) -> int:
        return self.width

    def get_height(self) -> int:
        return self.height

    def _set_width(self, width: int):
        self.width = width

    def _set_height(self, height: int):
        self.height = height

    def read(self) -> Tuple[bool, Frame]:
        return False, None
